# wifi_commands.py/Open GoPro, Version 2.0 (C) Copyright 2021 GoPro, Inc. (http:/gopro.com/OpenGoPro).
# This copyright was auto-generated on Tue Sep  7 21:35:52 UTC 2021

"""WiFi API for Open GoPro version 1.0"""

from __future__ import annotations
import logging
import datetime
from pathlib import Path
from typing import Any, Optional, Tuple, Dict

from open_gopro.communication_client import GoProWifi
from open_gopro.constants import SettingId, StatusId
from open_gopro.responses import BytesParserBuilder
from open_gopro.api.builders import WifiGetJsonNoParams, WifiGetJsonWithParams, WifiGetBinary, WifiSetting
from . import params as Params

logger = logging.getLogger(__name__)


class WifiCommands:
    """All of the Wifi commands.

    To be used as a delegate for a GoProWifi to build commands

    All of these return a GoProResp
    """

    # pylint: disable = missing-class-docstring, arguments-differ, useless-super-delegation
    def __init__(self, communicator: GoProWifi):
        """Constructor

        Args:
            communicator (GoProWifi):  Adapter to read / write commands
        """
        self.communicator = communicator

        class CameraFileToLocalFile(WifiGetBinary):
            def __call__(self, /, camera_file: str, local_file: Optional[Path] = None) -> Path:
                return super().__call__(camera_file=camera_file, local_file=local_file or camera_file)

        # ======================================== Commands

        self.set_digital_zoom = WifiGetJsonWithParams[int](
            communicator, "gopro/camera/digital_zoom?percent={}"
        )
        """Set digital zoom in percent.

        Args:
            value (int): Desired zoom as a percentage

        Returns:
            GoProResp: command status
        """

        self.get_camera_state = WifiGetJsonNoParams(communicator, "gopro/camera/state", parse_camera_state)
        """Get camera status and settings.

        Returns:
            GoProResp: command status
        """

        self.set_keep_alive = WifiGetJsonNoParams(communicator, "gopro/camera/keep_alive")
        """Send the keep alive signal to maintain the connection.

        Returns:
            GoProResp: command status
        """

        self.get_gpmf_data = CameraFileToLocalFile(communicator, "gopro/media/gpmf?path=100GOPRO/{}")
        """Get GPMF data for a file.

        If local_file is none, the output location will be the same name as the camera_file.

        Args:
            camera_file (str): filename on camera to operate on
            local_file (Optional[Path], optional): Location on computer to write output. Defaults to None.

        Returns:
            Path: Path to local_file that output was written to
        """

        self.get_media_info = WifiGetJsonWithParams[str](communicator, "gopro/media/info?path=100GOPRO/{}")
        """Get media info for a file.

        Args:
            value (str): Media file to get info for

        Returns:
            GoProResp: command status and media info as JSON
        """

        self.get_media_list = WifiGetJsonNoParams(
            communicator, "gopro/media/list", lambda x, _: {"files": x["media"][0]["fs"]}  # type: ignore
        )
        """Get a list of media on the camera.

        Returns:
            GoProResp: command status and extracted media list as JSON
        """

        self.get_screennail = CameraFileToLocalFile(communicator, "gopro/media/screennail?path=100GOPRO/{}")
        """Get screennail for a file.

        If local_file is none, the output location will be the same name as the camera_file.

        Args:
            camera_file (str): filename on camera to operate on
            local_file (Optional[Path], optional): Location on computer to write output. Defaults to None.

        Returns:
            Path: Path to local_file that output was written to
        """

        self.get_thumbnail = CameraFileToLocalFile(communicator, "gopro/media/thumbnail?path=100GOPRO/{}")
        """Get thumbnail for a file.

        If local_file is none, the output location will be the same name as the camera_file.

        Args:
            camera_file (str): filename on camera to operate on
            local_file (Optional[Path], optional): Location on computer to write output. Defaults to None.

        Returns:
            Path: Path to local_file that output was written to
        """

        self.set_turbo_mode = WifiGetJsonWithParams[Params.Toggle](
            communicator, "gopro/media/turbo_transfer?p={}"
        )
        """Enable or disable Turbo transfer mode.

        Args:
            value (Params.Toggle): enable / disable turbo mode

        Returns:
            GoProResp: command status
        """

        self.get_open_gopro_api_version = WifiGetJsonNoParams(communicator, "gopro/version")
        """Get Open GoPro version.

        Returns:
            GoProResp: command status and version separated into major and minor fields
        """

        self.get_preset_status = WifiGetJsonNoParams(communicator, "gopro/camera/presets/get")
        """Get current Preset status.

        Returns:
            GoProResp: command status and current preset status as JSON
        """

        self.set_preset = WifiGetJsonWithParams[Params.Preset](communicator, "gopro/camera/presets/load?id={}")
        """Set the active preset."""

        self.set_preset_group = WifiGetJsonWithParams[Params.PresetGroup](
            communicator, "gopro/camera/presets/set_group?id={}"
        )
        """Set the active preset group.

        The most recently used Preset in this group will be set.

        Args:
            value (Params.PresetGroup): desired Preset Group

        Returns:
            GoProResp: command status
        """

        self.start_preview_stream = WifiGetJsonNoParams(communicator, "gopro/camera/stream/start")
        """Start the preview stream.

        Returns:
            GoProResp: command status
        """

        self.stop_preview_stream = WifiGetJsonNoParams(communicator, "gopro/camera/stream/stop")
        """Stop the preview stream.

        Returns:
            GoProResp: command status
        """

        self.get_telemetry = CameraFileToLocalFile(communicator, "gopro/media/telemetry?path=100GOPRO/{}")
        """Download the telemetry data for a camera file and store in a local file.

        If local_file is none, the output location will be the same name as the camera_file.

        Args:
            camera_file (str): filename on camera to operate on
            local_file (Optional[Path], optional): Location on computer to write output. Defaults to None.

        Returns:
            Path: Path to local_file that output was written to
        """

        self.download_file = CameraFileToLocalFile(communicator, "videos/DCIM/100GOPRO/{}")
        """Download a video from the camera to a local file.

        If local_file is none, the output location will be the same name as the camera_file.

        Args:
            camera_file (str): filename on camera to operate on
            local_file (Optional[Path], optional): Location on computer to write output. Defaults to None.

        Returns:
            Path: Path to local_file that output was written to
        """

        self.set_third_party_client_info = WifiGetJsonNoParams(
            communicator, "gopro/camera/analytics/set_client_info"
        )
        """Flag as third party app.

        Will return error if the camera is already encoding.

        Returns:
            GoProResp: command status
        """

        self.set_shutter_on = WifiGetJsonNoParams(communicator, "gopro/camera/shutter/start")
        """Set the shutter on (i.e. start encoding).

        Returns:
            GoProResp: command status
        """

        self.set_shutter_off = WifiGetJsonNoParams(communicator, "gopro/camera/shutter/stop")
        """Set the shutter off (i.e. stop encoding).

        Will return error tf this command is sent when the camera is not encoding.

        Returns:
            GoProResp: command status
        """

        self.set_camera_control = WifiGetJsonWithParams[Params.CameraControl](
            communicator, "gopro/camera/control/set_ui_controller?p={}"
        )
        """Configure global behaviors by setting camera control (to i.e. Idle, External)

        Args:
            value (Params.CameraControl): desired camera control value

        Returns:
            GoProResp: command status
        """

        self.set_date_time = WifiGetJsonWithParams[datetime.datetime](
            communicator,
            "gopro/camera/set_date_time?{}",
            param_builder=lambda x: f"date={x.year}_{x.month}_{x.day}&time={x.hour}_{x.minute}_{x.second}",
        )
        """Update the date and time of the camera.

        Args:
            value (datetime.datetime): Time to set

        Returns:
            GoProResp: command status
        """

        self.get_date_time = WifiGetJsonNoParams(communicator, "gopro/camera/get_date_time")
        """Read the current date and time of the camera.

        Returns:
            GoProResp: command status that includes datetime.datetime
        """

        self.get_webcam_status = WifiGetJsonNoParams(communicator, "gopro/webcam/status")
        """Get the status of the webcam endpoint.

        Returns:
            GoProResp: command status that includes webcam status
        """

        self.get_webcam_version = WifiGetJsonNoParams(communicator, "gopro/webcam/version")
        """Get the version information of the webcam implementation.

        Returns:
            GoProResp: command status that includes the webcam version
        """

        self.add_photo_hilight = WifiGetJsonWithParams[str](
            communicator, "gopro/media/hilight/file?path=100GOPRO/{}"
        )
        """Add a hilight to a photo file (.jpg)

        Args:
            value (str): File to add the hilight to

        Returns:
            GoProResp: command status
        """

        self.add_video_hilight = WifiGetJsonWithParams[Tuple[str, int]](
            communicator, "gopro/media/hilight/file?path=100GOPRO/{}&ms={}"
        )
        """Add a hilight to a video file (.mp4)

        Args:
            value (Tuple[str, int]): (file, offset) where:
                \n\t- file is the video to remove the hilight
                \n\t- offset is the offset in ms

        Returns:
            GoProResp: command status
        """

        self.remove_photo_hilight = WifiGetJsonWithParams[str](
            communicator, "gopro/media/hilight/remove?path=100GOPRO/{}"
        )
        """Remove a hilight from a photo file (.jpg)

        Args:
            value (str): File to remove the hilight to

        Returns:
            GoProResp: command status
        """

        self.remove_video_hilight = WifiGetJsonWithParams[Tuple[str, int]](
            communicator, "gopro/media/hilight/remove?path=100GOPRO/{}&ms={}"
        )
        """Remove a hilight from a video file (.mp4)

        Args:
            value (Tuple[str, int]): (file, offset) where:
                \n\t- file is the video to remove the hilight
                \n\t- offset is the offset in ms

        Returns:
            GoProResp: command status
        """


# =====================================================Settings============================================


class WifiSettings:
    # pylint: disable=missing-class-docstring, unused-argument
    """Iterator to iterate through a BleSettings instance's attributes

    Does not include the 'communicator' instance.
    """

    class Iterator:
        """Iterator to iterate through a WifiSettings instance's attributes

        Does not include the 'communicator' instance.
        """

        def __init__(self, settings: "WifiSettings"):
            """Constructor

            Args:
                settings (WifiSettings): Settings to iterate through
            """
            self._index = 0
            self._setting_attributes = list(settings.__dict__.values())[1:]  # Skip communicator

        def __next__(self) -> WifiSetting:
            """Return the next attribute

            Raises:
                StopIteration: Nothing left to iterate

            Returns:
                WifiSetting: Instance of setting
            """
            if self._index < len(self._setting_attributes):
                setting = self._setting_attributes[self._index]
                self._index += 1
                return setting
            # End of Iteration
            raise StopIteration

    def __init__(self, communicator: GoProWifi, endpoint: str = "gopro/camera/setting?setting={}&option={}"):
        """The collection of all WiFi Settings

        Args:
            communicator (GoProWifi): Adapter to read / write settings
            endpoint (str): HTTP endpoint for command. Defaults to "gopro/camera/setting?setting={}&option={}".
        """
        self.endpoint = endpoint

        self.resolution = WifiSetting[Params.Resolution](communicator, SettingId.RESOLUTION)
        """Resolution. Set with :py:class:`open_gopro.Params.Resolution`"""

        self.fps = WifiSetting[Params.FPS](communicator, SettingId.FPS)
        """Frames per second. Set with :py:class:`open_gopro.Params.FPS`"""

        self.auto_off = WifiSetting[Params.AutoOff](communicator, SettingId.AUTO_OFF)
        """Set the auto off time. Set with :py:class:`Params.AutoOff`"""

        self.video_field_of_view = WifiSetting[Params.VideoFOV](communicator, SettingId.VIDEO_FOV)
        """Video FOV. Set with :py:class:`open_gopro.Params.FieldOfView`"""

        self.photo_field_of_view = WifiSetting[Params.PhotoFOV](communicator, SettingId.PHOTO_FOV)
        """Photo FOV. Set with :py:class:`open_gopro.Params.FieldOfView`"""

        self.multi_shot_field_of_view = WifiSetting[Params.MultishotFOV](
            communicator, SettingId.MULTI_SHOT_FOV
        )
        """Multi-shot FOV. Set with :py:class:`open_gopro.Params.FieldOfView`"""

        self.max_lens_mode = WifiSetting[Params.MaxLensMode](communicator, SettingId.MAX_LENS_MOD)
        """Enable / disable max lens mod. Set with :py:class:`open_gopro.Params.MaxLensMode`"""

        self.hypersmooth = WifiSetting[Params.HypersmoothMode](communicator, SettingId.HYPERSMOOTH)
        """Set / disable hypersmooth. Set with :py:class:`Params.HypersmoothMode`"""

        self.video_performance_mode = WifiSetting[Params.PerformanceMode](
            communicator, SettingId.VIDEO_PERFORMANCE_MODE
        )
        """Video Performance Mode (extended battery, tripod, etc)."""

        self.media_format = WifiSetting[Params.MediaFormat](communicator, SettingId.MEDIA_FORMAT)
        """Set the media format."""

    def __iter__(self) -> Iterator:
        """Return an iterable of this instance's attributes

        Does not include the 'communicator' attribute

        Returns:
            Iterator: next attribute
        """
        return WifiSettings.Iterator(self)


# ========================================= COMPLEX JSON RESPONSE PARSERS ====================================


def parse_camera_state(
    buf: Dict[str, Any], additional_parsers: Dict[Any, BytesParserBuilder] = None
) -> Dict[Any, Any]:
    """Parse the raw state values into user friendly types / values.

    Args:
        buf (Dict[str, Any]): input dict to parse
        additional_parsers (Dict[Any, BytesParserBuilder], optional): additional parsers to be used. Defaults to None

    Returns:
        Dict[Any, Any]: parsed output dict
    """
    assert additional_parsers is not None
    parsed: Dict[Any, Any] = {}
    # Parse status and settings values into nice human readable things
    for (name, id_map) in [("status", StatusId), ("settings", SettingId)]:
        for k, v in buf[name].items():
            identifier = id_map(int(k))
            try:
                parser = additional_parsers[identifier]
                val = parser.parse(parser.build(v))
            except KeyError:
                logger.warning(f"unparsed {name}: {identifier}")
                val = v
            except ValueError:
                logger.warning(f"{identifier} does not contain a value {v}")
                val = v

            parsed[identifier] = val

    return parsed



# __init__.py/Open GoPro, Version 2.0 (C) Copyright 2021 GoPro, Inc. (http://gopro.com/OpenGoPro).
# This copyright was auto-generated on Tue Sep  7 21:35:53 UTC 2021

"""Open GoPro WiFi Interface interace and implementation"""

from .controller import SsidState, WifiController
from .client import WifiClient



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['open_gopro',
 'open_gopro.api',
 'open_gopro.ble',
 'open_gopro.ble.adapters',
 'open_gopro.demos',
 'open_gopro.proto',
 'open_gopro.wifi',
 'open_gopro.wifi.adapters']

package_data = \
{'': ['*']}

install_requires = \
['betterproto[compiler]',
 'bleak==0.14.2',
 'construct',
 'packaging>=21.3,<22.0',
 'protobuf',
 'pydocstyle>=6.1.1,<7.0.0',
 'requests',
 'rich',
 'typing-extensions',
 'wrapt']

entry_points = \
{'console_scripts': ['gopro-ble-write = open_gopro.demos.ble_write:main',
                     'gopro-demo = open_gopro.demos.demo:main',
                     'gopro-log-battery = open_gopro.demos.log_battery:main',
                     'gopro-photo = open_gopro.demos.photo:entrypoint',
                     'gopro-stream = open_gopro.demos.stream:main',
                     'gopro-video = open_gopro.demos.video:main',
                     'gopro-wifi = open_gopro.demos.connect_wifi:main']}

setup_kwargs = {
    'name': 'open-gopro',
    'version': '0.9.1',
    'description': 'Open GoPro API and Examples',
    'long_description': '# Open GoPro Python SDK\n\n<img alt="GoPro Logo" src="https://raw.githubusercontent.com/gopro/OpenGoPro/main/docs/assets/images/logos/logo.png" width="50%" style="max-width: 500px;"/>\n\n[![Build and Test](https://img.shields.io/github/workflow/status/gopro/OpenGoPro/Python%20SDK%20Testing?label=Build%20and%20Test)](https://github.com/gopro/OpenGoPro/actions/workflows/python_sdk_test.yml)\n[![Build Docs](https://img.shields.io/github/workflow/status/gopro/OpenGoPro/Python%20SDK%20Docs%20Build%20and%20Deploy?label=Docs)](https://github.com/gopro/OpenGoPro/actions/workflows/python_sdk_deploy_docs.yml)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n[![PyPI](https://img.shields.io/pypi/v/open-gopro)](https://pypi.org/project/open-gopro/)\n[![MIT License](https://img.shields.io/badge/License-MIT-blue.svg)](https://github.com/gopro/OpenGoPro/blob/main/LICENSE)\n![Coverage](https://raw.githubusercontent.com/gopro/OpenGoPro/main/demos/python/sdk_wireless_camera_control/docs/_static/coverage.svg)\n\nThis is a Python package that provides an interface for the user to exercise the Open GoPro Bluetooth Low\nEnergy (BLE) and Wi-Fi API\'s as well as install command line interfaces to take photos, videos, and view\nthe preview stream.\n\n-   Free software: MIT license\n-   Documentation: [View on Open GoPro](https://gopro.github.io/OpenGoPro/python_sdk/)\n-   View on [Github](https://github.com/gopro/OpenGoPro/tree/main/demos/python/sdk_wireless_camera_control)\n\n## Documentation\n\n> Note! This README is only an overview of the package.\n\nComplete documentation can be found on [Open GoPro](https://gopro.github.io/OpenGoPro/python_sdk/)\n\n## Features\n\n-   Top-level GoPro class interface to use both BLE / WiFi\n-   Cross-platform (tested on MacOS Big Sur, Windows 10, and Ubuntu 20.04)\n    -   BLE implemented using [bleak](https://pypi.org/project/bleak/)\n    -   Wi-Fi controller provided in the Open GoPro package (loosely based on the [Wireless Library](https://pypi.org/project/wireless/)\n-   Supports all commands, settings, and statuses from the [Open GoPro API](https://gopro.github.io/OpenGoPro/)\n-   Supports all versions of the Open GoPro API\n-   Automatically handles connection maintenance:\n    -   manage camera ready / encoding\n    -   periodically sends keep alive signals\n-   Includes detailed logging for each module\n-   Includes demo scripts installed as command-line applications to show BLE and WiFi functionality\n    -   Take a photo\n    -   Take a video\n    -   View the live stream\n    -   Log the battery\n\n## Installation\n\n> Note! This package requires Python >= 3.8\n\n```console\n    $ pip install open-gopro\n```\n\n## Usage\n\nTo automatically connect to GoPro device via BLE and WiFI, set the preset, set video parameters, take a\nvideo, and download all files:\n\n```python\nimport time\nfrom open_gopro import GoPro, Params\n\nwith GoPro() as gopro:\n    gopro.ble_command.load_preset(Params.Preset.CINEMATIC)\n    gopro.ble_setting.resolution.set(Params.Resolution.RES_4K)\n    gopro.ble_setting.fps.set(Params.FPS.FPS_30)\n    gopro.ble_command.set_shutter(Params.Shutter.ON)\n    time.sleep(2) # Record for 2 seconds\n    gopro.ble_command.set_shutter(Params.Shutter.OFF)\n\n    # Download all of the files from the camera\n    media_list = [x["n"] for x in gopro.wifi_command.get_media_list().flatten\n    for file in media_list:\n        gopro.wifi_command.download_file(camera_file=file)\n```\n\nAnd much more!\n\n## Demos\n\n> Note! These demos can be found on [Github](https://github.com/gopro/OpenGoPro/tree/main/demos/python/sdk_wireless_camera_control/open_gopro/demos)\n\nDemos can be found in the installed package in the "demos" folder. They are installed as a CLI entrypoint\nand can be run via:\n\nCapture a photo and download it to your computer:\n\n```bash\n$ gopro-photo\n```\n\nCapture a video and download it to your computer:\n\n```bash\n$ gopro-video\n```\n\nStart the preview stream and view it with [VLC](https://www.videolan.org/):\n\n```bash\n$ gopro-stream\n```\n\nConnect to the GoPro and log battery consumption in to a .csv:\n\n```bash\n$ gopro-log-battery\n```\n\nConnect to the GoPro\'s Wi-Fi AP and maintain the connection:\n\n```bash\n$ gopro-wifi\n```\n\nFor more information on each, try running with help as such:\n\n```bash\n$ gopro-photo --help\n\nusage: gopro-photo [-h] [-i IDENTIFIER] [-l LOG] [-o OUTPUT] [-w WIFI_INTERFACE]\n\nConnect to a GoPro camera, take a photo, then download it.\n\noptional arguments:\n  -h, --help            show this help message and exit\n  -i IDENTIFIER, --identifier IDENTIFIER\n                        Last 4 digits of GoPro serial number, which is the last 4 digits of the default camera SSID. If not used, first\n                        discovered GoPro will be connected to\n  -l LOG, --log LOG     Location to store detailed log\n  -o OUTPUT, --output OUTPUT\n                        Where to write the photo to. If not set, write to \'photo.jpg\'\n  -w WIFI_INTERFACE, --wifi_interface WIFI_INTERFACE\n                        System Wifi Interface. If not set, first discovered interface will be used.\n```\n',
    'author': 'Tim Camise',
    'author_email': 'tcamise@gopro.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/gopro/OpenGoPro/tree/main/demos/python/sdk_wireless_camera_control',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



# util.py/Open GoPro, Version 2.0 (C) Copyright 2021 GoPro, Inc. (http://gopro.com/OpenGoPro).
# This copyright was auto-generated on Wed, Sep  1, 2021  5:05:50 PM

"""Miscellaneous utilities for the GoPro package."""

from __future__ import annotations
import sys
import enum
import json
import types
import queue
import logging
import subprocess
import dataclasses
from pathlib import Path
from base64 import b64encode
import http.client as http_client
from datetime import timedelta, datetime
from typing import Dict, Type, Any, List, Optional, Union

import betterproto
from rich.logging import RichHandler
from rich import traceback

util_logger = logging.getLogger(__name__)

# From https://stackoverflow.com/questions/2183233/how-to-add-a-custom-loglevel-to-pythons-logging-facility/35804945#35804945
def addLoggingLevel(levelName: str, levelNum: int) -> None:
    """Comprehensively adds a new logging level to the `logging` module and the currently configured logging class.

    `levelName` becomes an attribute of the `logging` module with the value
    `levelNum`. `methodName` becomes a convenience method for both `logging`
    itself and the class returned by `logging.getLoggerClass()` (usually just
    `logging.Logger`). If `methodName` is not specified, `levelName.lower()` is
    used.

    To avoid accidental clobberings of existing attributes, this method will
    raise an `AttributeError` if the level name is already an attribute of the
    `logging` module or if the method name is already present

    Example:
    --------
    >>> addLoggingLevel('TRACE', logging.DEBUG - 5)
    >>> logging.getLogger(__name__).setLevel("TRACE")
    >>> logging.getLogger(__name__).trace('that worked')
    >>> logging.trace('so did this')
    >>> logging.TRACE
    5

    Args:
        levelName (str): name of level (i.e. TRACE)
        levelNum (int): integer level of new logging level
    """
    methodName = levelName.lower()

    def logForLevel(self: Any, message: str, *args: Any, **kwargs: Any) -> None:
        if self.isEnabledFor(levelNum):
            self._log(levelNum, message, args, **kwargs)

    def logToRoot(message: str, *args: Any, **kwargs: Any) -> None:
        logging.log(levelNum, message, *args, **kwargs)

    logging.addLevelName(levelNum, levelName)
    setattr(logging, levelName, levelNum)
    setattr(logging.getLoggerClass(), methodName, logForLevel)
    setattr(logging, methodName, logToRoot)


def setup_logging(
    logger: Any, output: Optional[Path] = None, modules: Dict[str, int] = None
) -> logging.Logger:
    """Configure open gopro modules for logging

    The application's logger is passed in, modified, and then returned

    Args:
        logger (Any): input logger that will be modified and then returned
        output (Path, optional): Path of log file for file stream handler. If not set, will not log to file.
        modules (Dict[str, int], optional): Optional override of modules / levels. Will be merged into default
            modules.

    Returns:
        Any: updated logger that the application can use for logging
    """
    default_modules = {
        "open_gopro.gopro": logging.DEBUG,
        "open_gopro.api.builders": logging.DEBUG,
        "open_gopro.api.wifi_commands": logging.DEBUG,
        "open_gopro.api.ble_commands": logging.DEBUG,
        "open_gopro.communication_client": logging.DEBUG,
        "open_gopro.ble.adapters.bleak_wrapper": logging.INFO,
        "open_gopro.ble.client": logging.DEBUG,
        "open_gopro.wifi.adapters.wireless": logging.DEBUG,
        "open_gopro.responses": logging.DEBUG,
        "open_gopro.util": logging.DEBUG,
        "bleak": logging.WARNING,
        "urllib3": logging.WARNING,
        "http.client": logging.WARNING,
    }

    logging_modules = {**default_modules, **modules} if modules else default_modules

    # monkey-patch a `print` global into the http.client module; all calls to
    # print() in that module will then use our logger's debug method
    http_client.HTTPConnection.debuglevel = 1
    http_client.print = lambda *args: logging.getLogger("http.client").debug(" ".join(args))  # type: ignore

    addLoggingLevel("TRACE", logging.DEBUG - 5)

    if output:
        # Logging to file with millisecond timing
        fh = logging.FileHandler(output, mode="w")
        file_formatter = logging.Formatter(
            fmt="%(threadName)13s:%(asctime)s.%(msecs)03d %(filename)-40s %(lineno)4s %(levelname)-8s | %(message)s",
            datefmt="%H:%M:%S",
        )
        fh.setFormatter(file_formatter)
        fh.setLevel(logging.TRACE)  # type: ignore # pylint: disable=no-member
        logger.addHandler(fh)
    else:
        fh = None

    # Use Rich for colorful console logging
    sh = RichHandler(rich_tracebacks=True, enable_link_path=True, show_time=False)
    stream_formatter = logging.Formatter("%(asctime)s.%(msecs)03d %(message)s", datefmt="%H:%M:%S")
    sh.setFormatter(stream_formatter)
    sh.setLevel(logging.INFO)
    logger.addHandler(sh)
    logger.setLevel(logging.TRACE)  # type: ignore # pylint: disable=no-member

    # Enable / disable logging in modules
    for module, level in logging_modules.items():
        l = logging.getLogger(module)
        l.setLevel(level)
        if fh:
            l.addHandler(fh)
        l.addHandler(sh)

    traceback.install()  # Enable exception tracebacks in rich logger

    return logger


def set_logging_level(logger: Any, level: int) -> None:
    """Change the global logging level

    Args:
        logger (Any): logger to update
        level (int): level to set
    """
    for handler in logger.handlers:
        if isinstance(handler, RichHandler):
            handler.setLevel(level)


ARROW_HEAD_COUNT = 8
ARROW_TAIL_COUNT = 14


def build_log_tx_str(stringable: Any) -> str:
    """Build a string with Tx arrows

    Args:
        stringable (Any): stringable object to surround with arrows

    Returns:
        str: string surrounded by Tx arrows
    """
    arrow = f"{'<'*ARROW_HEAD_COUNT}{'-'*ARROW_TAIL_COUNT}"
    return f"\n\n{arrow}\n\t{stringable}\n{arrow}\n"


def build_log_rx_str(stringable: Any, asynchronous: bool = False) -> str:
    """Build a string with Rx arrows

    Args:
        stringable (Any): stringable object to surround with arrows
        asynchronous (bool): Should the arrows contain ASYNC?. Defaults to False.

    Returns:
        str: string surrounded by Rx arrows
    """
    assert ARROW_TAIL_COUNT > 5
    if asynchronous:
        arrow = f"{'-'*(ARROW_TAIL_COUNT//2-3)}ASYNC{'-'*(ARROW_TAIL_COUNT//2-2)}{'>'*ARROW_HEAD_COUNT}"
    else:
        arrow = f"{'-'*ARROW_TAIL_COUNT}{'>'*ARROW_HEAD_COUNT}"
    return f"\n\n{arrow}\n{stringable}\n{arrow}\n"


def launch_vlc(location: Optional[Path]) -> None:
    """Launch VLC

    Args:
        location (Optional[Path]): path to VLC. If None, it will be automatically discovered
    """
    # This is a fairly lazy way to find VLC. We'll call it best effort.
    potential_vlc_locations: List[Union[Path, str]] = []
    command = "echo Invalid Platform"
    if "linux" in sys.platform.lower():
        potential_vlc_locations = [r'"/snap/bin/vlc"']
        command = 'su $(id -un 1000) -c "{} udp://@:8554 > /dev/null 2>&1 &"'
    elif "darwin" in sys.platform.lower():
        potential_vlc_locations = [r'"/Applications/VLC.app/Contents/MacOS/VLC"']
        command = "{} udp://@:8554 > /dev/null 2>&1 &"
    elif "win" in sys.platform.lower():
        potential_vlc_locations = [
            r'"/c/Program Files/VideoLAN/VLC/vlc.exe"',
            r'"/c/Program Files (x86)/VideoLAN/VLC/vlc.exe"',
            r'"C:\Program Files (x86)\VideoLAN\VLC\vlc.exe"',
            r'"C:\Program Files\VideoLAN\VLC\vlc.exe"',
        ]
        command = "{} udp://@:8554 &"

    potential_vlc_locations = potential_vlc_locations if location is None else [location]
    for vlc in potential_vlc_locations:
        response = cmd(command.format(vlc)).lower()

        if (
            " not " not in response
            and " no " not in response
            and " cannot " not in response
            and " unexpected " not in response
        ):
            util_logger.info("VLC launched")
            return

    util_logger.error("Failed to find VLC")


def scrub(obj: Any, bad_key: str) -> None:
    """Recursively scrub a dict or list to remove a given key in place.

    Args:
        obj (Any): dict or list to operate on. If neither, it will return immediately.
        bad_key (str): key to remove
    """
    if isinstance(obj, dict):
        for key in list(obj.keys()):
            if key == bad_key:
                del obj[key]
            else:
                scrub(obj[key], bad_key)
    elif isinstance(obj, list):
        for i in reversed(range(len(obj))):
            if obj[i] == bad_key:
                del obj[i]
            else:
                scrub(obj[i], bad_key)
    else:
        # neither a dict nor a list, do nothing
        pass


def pretty_print(obj: Any) -> str:
    """Recursively iterate through object and turn elements into strings as desired for eventual json dumping

    Args:
        obj (Any): object to recurse through

    Returns:
        str: pretty-printed string
    """

    def stringify(elem: Any) -> Union[str, int, float]:
        """Get the string value of an element if it is not a number (int, float, etc.)

        Special case for IntEnum since json refuses to treat these as strings.

        Args:
            elem (Any): element to potentially stringify

        Returns:
            str: string representation
        """
        if isinstance(elem, (bytes, bytearray)):
            return elem.hex(":")
        if isinstance(elem, enum.Enum) and isinstance(elem, int):
            return str(elem)
        if not isinstance(elem, (int, float)):
            return str(elem)
        return elem

    def recurse(elem: Any) -> Any:
        """Recursion function

        Args:
            elem (Any): current element to work on

        Returns:
            Any: element after recursion is done
        """
        if isinstance(elem, dict):
            # nested dictionary
            new_dic = {}
            for k, v in elem.items():
                if isinstance(v, (dict, list)):
                    new_dic[recurse(k)] = recurse(v)
                else:
                    new_dic[recurse(k)] = stringify(v)

            return new_dic

        if isinstance(elem, list):
            # nested list
            return [recurse(t) for t in elem]

        return stringify(elem)

    return json.dumps(recurse(obj), indent=4)


def cmd(command: str) -> str:
    """Send a command to the shell and return the result.

    Args:
        command (str): command to send

    Returns:
        str: response returned from shell
    """
    # We don't want password showing in the log
    if "sudo" in command:
        logged_command = command[: command.find('"') + 1] + "********" + command[command.find(" | sudo") - 1 :]
    else:
        logged_command = command
    util_logger.debug(f"Send cmd --> {logged_command}")
    # Note: Ignoring unicode characters in SSIDs to prevent intermittent UnicodeDecodeErrors from occurring
    # while trying to connect to SSID when *any* AP is nearby that has unicode characters in the name
    response = (
        subprocess.Popen(command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)  # type: ignore
        .stdout.read()
        .decode(errors="ignore")
    )
    util_logger.debug(f"Receive response --> {response}")

    return response


class Singleton:
    """To be subclassed to create a singleton class."""

    _instances: Dict[Type[Singleton], Singleton] = {}

    def __new__(cls, *_: Any) -> Any:  # noqa https://github.com/PyCQA/pydocstyle/issues/515
        if cls not in cls._instances:
            cls._instances[cls] = object.__new__(cls)
        return cls._instances[cls]


class SnapshotQueue(queue.Queue):
    """A subclass of the default queue module to safely take a snapshot of the queue

    This is so we can access the elements (in a thread safe manner) without dequeuing them.
    """

    def snapshot(self) -> List[Any]:
        """Acquire the mutex, then return the queue's elements as a list.

        Returns:
            List[Any]: List of queue elements
        """
        with self.mutex:
            return list(self.queue)


# ============================================================================================================


def custom_betterproto_to_dict(
    self: betterproto.Message,
    casing: betterproto.Casing = betterproto.Casing.CAMEL,
    include_default_values: bool = False,
) -> dict:
    """TODO

    Args:
        self (betterproto.Message): [description]
        casing (betterproto.Casing, optional): [description]. Defaults to betterproto.Casing.CAMEL.
        include_default_values (bool): [description]. Defaults to False.

    Returns:
        dict: [description]
    """
    output: Dict[str, Any] = {}
    for field in dataclasses.fields(self):
        meta = betterproto.FieldMetadata.get(field)
        v = getattr(self, field.name)
        cased_name = casing(field.name).rstrip("_")  # type: ignore
        if meta.proto_type == "message":
            if isinstance(v, datetime):
                if v != betterproto.DATETIME_ZERO or include_default_values:
                    output[cased_name] = betterproto._Timestamp.timestamp_to_json(v)
            elif isinstance(v, timedelta):
                if v != timedelta(0) or include_default_values:
                    output[cased_name] = betterproto._Duration.delta_to_json(v)
            elif meta.wraps:
                if v is not None or include_default_values:
                    output[cased_name] = v
            elif isinstance(v, list):
                # Convert each item.
                values = []
                for i in v:
                    i.to_dict = types.MethodType(custom_betterproto_to_dict, i)
                    values.append(i.to_dict(casing, include_default_values))
                if values or include_default_values:
                    output[cased_name] = values
            else:
                if v._serialized_on_wire or include_default_values:
                    v.to_dict = types.MethodType(custom_betterproto_to_dict, v)
                    output[cased_name] = v.to_dict(casing, include_default_values)
        elif meta.proto_type == "map":
            for k in v:
                if hasattr(v[k], "to_dict"):
                    v.to_dict = types.MethodType(custom_betterproto_to_dict, v)
                    v[k] = v[k].to_dict(casing, include_default_values)

            if v or include_default_values:
                output[cased_name] = v
        elif v != self._get_field_default(field, meta) or include_default_values:
            if meta.proto_type in betterproto.INT_64_TYPES:
                if isinstance(v, list):
                    output[cased_name] = [str(n) for n in v]
                else:
                    output[cased_name] = str(v)
            elif meta.proto_type == betterproto.TYPE_BYTES:
                if isinstance(v, list):
                    output[cased_name] = [b64encode(b).decode("utf8") for b in v]
                else:
                    output[cased_name] = b64encode(v).decode("utf8")
            elif meta.proto_type == betterproto.TYPE_ENUM:
                enum_values = {}
                for e in self._betterproto.cls._cls_for(field):
                    enum_values[e.value] = e
                if isinstance(v, list):
                    output[cased_name] = [enum_values[e] for e in v]
                else:
                    output[cased_name] = enum_values[v]
            else:
                output[cased_name] = v
    return output


def build_protos() -> None:
    """Build the protobuf source .py files from the .proto files

    This is meant to be the entrypoint for the poe task
    """
    current_dir = Path(__file__).parent.resolve()
    proto_src_dir = current_dir / ".." / ".." / ".." / ".." / "protobuf"
    proto_out_dir = current_dir / "proto"
    print(f"current dir: {current_dir}")
    for file in proto_src_dir.iterdir():
        proto_out = f"{file.name.split('.')[0]}_pb.py"
        print(f"building {proto_out} from {file.name} ...")
        cmd(f"mv {proto_out_dir / '__init__.py'} {proto_out_dir / 'temp'}")
        cmd(
            f"poetry run python -m grpc_tools.protoc  -I {proto_src_dir} --python_betterproto_out={proto_out_dir} {file}"
        )
        cmd(f"mv {proto_out_dir / 'open_gopro.py'} {proto_out_dir / proto_out}")
        cmd(f"mv {proto_out_dir / 'temp'} {proto_out_dir / '__init__.py'}")
